/* FreeRTOS */
#include "FreeRTOS.h"
#include "semphr.h"
#include "task.h"

/* HAL */
#include "stm32l4xx_hal.h"

/* MX */
#include "dma.h"
#include "gpio.h"
#include "spi.h"

/* LCD */
#include "epd-spi/driver/lcd_st7789.h"

#define SET_DC(x)  HAL_GPIO_WritePin(LCD_DC_GPIO_Port, LCD_DC_Pin, x)
#define SET_RST(x) HAL_GPIO_WritePin(LCD_RST_GPIO_Port, LCD_RST_Pin, x)
#define SET_BLK(x) HAL_GPIO_WritePin(LCD_BLK_GPIO_Port, LCD_BLK_Pin, x)

static SemaphoreHandle_t s_spi_dma_semaphore = NULL;

static void app_lcd_spi_evt_callback(SPI_HandleTypeDef *hspi);

int app_lcd_impl_init(void) {
    SET_BLK(GPIO_PIN_RESET);
    SET_DC(GPIO_PIN_SET);
    SET_RST(GPIO_PIN_SET);

    HAL_SPI_RegisterCallback(&hspi1, HAL_SPI_TX_COMPLETE_CB_ID, app_lcd_spi_evt_callback);

    s_spi_dma_semaphore = xSemaphoreCreateBinary();
    if (s_spi_dma_semaphore == NULL) {
        return -1;
    }

    return 0;
}

epd_ret_t app_lcd_impl_write_command(void *handle, uint8_t *command, uint32_t len) {
    HAL_StatusTypeDef ret;

    SET_DC(GPIO_PIN_RESET);
    ret = HAL_SPI_Transmit(&hspi1, command, 0x01, 1000);
    if (ret != HAL_OK) {
        return EPD_FAIL;
    }

    if (len > 1) {
        SET_DC(GPIO_PIN_SET);
        ret = HAL_SPI_Transmit(&hspi1, &command[1], len - 1, 1000);
        if (ret != HAL_OK) {
            return EPD_FAIL;
        }
    }

    return EPD_OK;
}

epd_ret_t app_lcd_impl_write_data(void *handle, const uint8_t *data, uint32_t len) {
    HAL_StatusTypeDef ret;

    SET_DC(GPIO_PIN_SET);
    ret = HAL_SPI_Transmit_DMA(&hspi1, (uint8_t *)data, len);
    if (ret != HAL_OK) {
        return EPD_FAIL;
    }

    xSemaphoreTake(s_spi_dma_semaphore, portMAX_DELAY);

    return EPD_OK;
}

epd_ret_t app_lcd_impl_reset(void *handle) {
    SET_RST(GPIO_PIN_RESET);
    vTaskDelay(pdMS_TO_TICKS(100));
    SET_RST(GPIO_PIN_SET);

    return EPD_OK;
}

epd_ret_t app_lcd_impl_delay(void *handle, uint32_t msec) {
    vTaskDelay(pdMS_TO_TICKS(msec));

    return EPD_OK;
}

epd_ret_t app_lcd_impl_backlight(void *handle, uint8_t on) {
    if (on) {
        SET_BLK(GPIO_PIN_SET);
    } else {
        SET_BLK(GPIO_PIN_RESET);
    }

    return EPD_OK;
}

static void app_lcd_spi_evt_callback(SPI_HandleTypeDef *hspi) {
    BaseType_t hp_woken = pdFALSE;
    xSemaphoreGiveFromISR(s_spi_dma_semaphore, &hp_woken);
    portYIELD_FROM_ISR(hp_woken);
}