#include <stdint.h>
#include <stdio.h>
#include <string.h>

/* FreeRTOS */
#include "FreeRTOS.h"
#include "queue.h"
#include "semphr.h"
#include "task.h"

/* HAL */
#include "stm32l4xx_hal.h"

/* MX */
#include "dma.h"

/* LCD */
#include "epd-spi/panel/lcd_h144c121d.h"

/* App */
#include "app_lcd_impl.h"
#include "illumination/LampArrayHidStructs.h"

#define APP_LAMP_COUNT        64
#define APP_LAMP_QUEUE_LENGTH 64

#define APP_LCD_BUF_WIDTH  30
#define APP_LCD_BUF_HEIGHT 30

typedef enum {
    APP_LAMP_UPDATE_SINGLE,
    APP_LAMP_UPDATE_MULTI,
} app_lighting_update_cmd_t;

typedef struct {
    app_lighting_update_cmd_t cmd;
    uint8_t                   lamp_id_start;
    uint8_t                   lamp_id_end;
    uint16_t                  color_565;
} app_lighting_q_item_t;

static uint16_t s_active_lamp_id = 0U;
static Position s_lamp_pos[APP_LAMP_COUNT];
static uint8_t  s_lamp_buf[APP_LCD_BUF_WIDTH * APP_LCD_BUF_HEIGHT * 2];

static QueueHandle_t     s_lamp_update_queue     = NULL;
static SemaphoreHandle_t s_memory_fill_semaphore = NULL;

static lcd_st7789_t s_lcd = {
    .cb =
        {
            .write_command_cb = app_lcd_impl_write_command,
            .write_data_cb    = app_lcd_impl_write_data,
            .reset_cb         = app_lcd_impl_reset,
            .backlight_cb     = app_lcd_impl_backlight,
            .delay_cb         = app_lcd_impl_delay,
        },
    .user_data = NULL,
};

static void app_lighting_fill_callback(DMA_HandleTypeDef *hdma);
static void app_lighting_control_task(void *parameters);

int app_lighting_init(void) {
    if (app_lcd_impl_init() < 0) {
        return -1;
    }

    if (lcd_st7789_init(&s_lcd, &lcd_h144c121d_panel_config) != EPD_OK) {
        return -2;
    }

    s_lamp_update_queue = xQueueCreate(APP_LAMP_QUEUE_LENGTH, sizeof(app_lighting_q_item_t));
    if (s_lamp_update_queue == NULL) {
        return -3;
    }

    s_memory_fill_semaphore = xSemaphoreCreateBinary();
    if (s_memory_fill_semaphore == NULL) {
        return -4;
    }

    for (uint8_t i = 0; i < APP_LAMP_COUNT; i++) {
        s_lamp_pos[i].x = (i % 8) * 10000;
        s_lamp_pos[i].y = (i / 8) * 10000;
        s_lamp_pos[i].z = 2500;
    }

    HAL_DMA_RegisterCallback(&hdma_memtomem_dma1_channel4, HAL_DMA_XFER_CPLT_CB_ID, app_lighting_fill_callback);

    if (xTaskCreate(app_lighting_control_task, "LC", 512, NULL, 3, NULL) != pdPASS) {
        return -5;
    }

    return 0;
}

uint16_t app_lighting_get_array_attributes(uint8_t *buffer) {
    LampArrayAttributesReport report = {
        .lampCount     = APP_LAMP_COUNT,
        .width         = 80000,
        .height        = 80000,
        .depth         = 5000,
        .lampArrayKind = 0x0A,
    };

    memcpy(buffer, &report, sizeof(LampArrayAttributesReport));
    return sizeof(report);
}

uint16_t app_lighting_get_lamp_attributes(uint8_t *buffer) {
    LampAttributesResponseReport report = {
        .lampId              = s_active_lamp_id,              // LampId
        .lampPosition        = s_lamp_pos[s_active_lamp_id],  // Lamp position
        .updateLatency       = 33000,                         // Lamp update interval
        .lampPurposes        = LAMP_PURPOSE_ACCENT,           // Lamp purpose
        .redLevelCount       = 255,                           // Red level count
        .blueLevelCount      = 255,                           // Blue level count
        .greenLevelCount     = 255,                           // Green level count
        .intensityLevelCount = 1,                             // Intensity
        .isProgrammable      = 1,                             // Is Programmable
    };

    memcpy(buffer, &report, sizeof(LampAttributesResponseReport));

    s_active_lamp_id++;
    if (s_active_lamp_id == APP_LAMP_COUNT) {
        s_active_lamp_id = 0;
    }

    return sizeof(report);
}

void app_lighting_set_lamp_attributes_req(const uint8_t *buffer) {
    LampAttributesRequestReport *report = (LampAttributesRequestReport *)buffer;

    s_active_lamp_id = report->lampId;
}

void app_lighting_set_lamp_multi_update(const uint8_t *buffer) {
    LampMultiUpdateReport *report = (LampMultiUpdateReport *)buffer;

    for (int i = 0; i < report->lampCount; i++) {
        uint16_t color_565 = 0U;
        color_565 |= ((report->colors[i].red >> 3) & 0x1F) << 11U;
        color_565 |= ((report->colors[i].green >> 2) & 0x3F) << 5U;
        color_565 |= ((report->colors[i].blue) >> 3) & 0x1F;

        app_lighting_q_item_t lc_qi = {
            .cmd           = APP_LAMP_UPDATE_SINGLE,
            .color_565     = color_565,
            .lamp_id_start = report->lampIds[i],
        };

        xQueueSend(s_lamp_update_queue, &lc_qi, pdMS_TO_TICKS(20));
    }
}

void app_lighting_set_lamp_range_update(const uint8_t *buffer) {
    LampRangeUpdateReport *report = (LampRangeUpdateReport *)buffer;

    uint16_t color_565 = 0U;
    color_565 |= ((report->color.red >> 3) & 0x1F) << 11U;
    color_565 |= ((report->color.green >> 2) & 0x3F) << 5U;
    color_565 |= ((report->color.blue) >> 3) & 0x1F;

    app_lighting_q_item_t lc_qi = {
        .cmd           = APP_LAMP_UPDATE_MULTI,
        .color_565     = color_565,
        .lamp_id_start = report->lampIdStart,
        .lamp_id_end   = report->lampIdEnd,
    };
    xQueueSend(s_lamp_update_queue, &lc_qi, pdMS_TO_TICKS(20));
}

void app_lighting_set_lamp_array_control(const uint8_t *buffer) {
}

static void app_lighting_fill_lcd_block(uint16_t color, uint8_t *buf) {
    uint16_t color_swapped = ((color & 0xFFU) << 8U) | (color >> 8U);

    HAL_DMA_Start_IT(&hdma_memtomem_dma1_channel4, (uint32_t)&color_swapped, (uint32_t)buf,
                     (APP_LCD_BUF_WIDTH * APP_LCD_BUF_HEIGHT));

    xSemaphoreTake(s_memory_fill_semaphore, portMAX_DELAY);
}

static void app_lighting_update_lamp(uint8_t id, const uint8_t *buf) {
    epd_coord_t coord = {
        .x_start = (id % 8) * APP_LCD_BUF_WIDTH,
        .x_end   = ((id % 8) * APP_LCD_BUF_WIDTH + APP_LCD_BUF_WIDTH - 1),
        .y_start = (id / 8) * APP_LCD_BUF_HEIGHT,
        .y_end   = ((id / 8) * APP_LCD_BUF_HEIGHT + APP_LCD_BUF_HEIGHT - 1),
    };

    lcd_st7789_load(&s_lcd, &coord, buf);
}

static void app_lighting_fill_callback(DMA_HandleTypeDef *hdma) {
    BaseType_t hp_woken = pdFALSE;
    xSemaphoreGiveFromISR(s_memory_fill_semaphore, &hp_woken);
    portYIELD_FROM_ISR(hp_woken);
}

static void app_lighting_control_task(void *parameters) {
    app_lighting_q_item_t lc_qi;

    lcd_st7789_enable_display(&s_lcd, true);

    for (;;) {
        xQueueReceive(s_lamp_update_queue, &lc_qi, portMAX_DELAY);
        app_lighting_fill_lcd_block(lc_qi.color_565, s_lamp_buf);

        HAL_GPIO_TogglePin(LED1_GPIO_Port, LED1_Pin);

        if (lc_qi.cmd == APP_LAMP_UPDATE_SINGLE) {
            app_lighting_update_lamp(lc_qi.lamp_id_start, s_lamp_buf);
        } else {
            for (int i = lc_qi.lamp_id_start; i < (lc_qi.lamp_id_end + 1); i++) {
                app_lighting_update_lamp(i, s_lamp_buf);
            }
        }
    }
}