#include "tusb.h"

/* HAL */
#include "stm32l4xx_hal.h"

/* App */
#include "app_lighting.h"
#include "app_tusb.h"
#include "app_tusb_descriptors.h"

static void app_tusb_task(void *parameters);
static int  app_tusb_peripheral_init(void);

int app_tusb_init(void) {
    if (app_tusb_peripheral_init() < 0) {
        return -1;
    }

    NVIC_SetPriority(OTG_FS_IRQn, configLIBRARY_MAX_SYSCALL_INTERRUPT_PRIORITY + 1);
    NVIC_EnableIRQ(OTG_FS_IRQn);

    if (tud_init(BOARD_TUD_RHPORT) == false) {
        return -2;
    }

    if (xTaskCreate(app_tusb_task, "TU", 1024, NULL, 4, NULL) != pdPASS) {
        return -3;
    }

    return 0;
}

void app_tusb_irq_handler(void) {
    tud_int_handler(BOARD_TUD_RHPORT);
}

static void app_tusb_task(void *parameters) {
    for (;;) {
        tud_task();
    }
}

static int app_tusb_peripheral_init(void) {
    RCC_OscInitTypeDef osc_init = {0};

    osc_init.OscillatorType = RCC_OSCILLATORTYPE_HSI48;
    osc_init.HSI48State     = RCC_HSI48_ON;
    if (HAL_RCC_OscConfig(&osc_init) != HAL_OK) {
        return -1;
    }

    RCC_PeriphCLKInitTypeDef peri_clk_init = {0};

    peri_clk_init.PeriphClockSelection = RCC_PERIPHCLK_USB;
    peri_clk_init.UsbClockSelection    = RCC_USBCLKSOURCE_HSI48;

    if (HAL_RCCEx_PeriphCLKConfig(&peri_clk_init) != HAL_OK) {
        return -2;
    }

    __HAL_RCC_USB_OTG_FS_CLK_ENABLE();
    HAL_PWREx_EnableVddUSB();

    GPIO_InitTypeDef gpio_init = {0};

    __HAL_RCC_GPIOA_CLK_ENABLE();
    __HAL_RCC_GPIOB_CLK_ENABLE();

    /* USB DP/DM pins */
    gpio_init.Pin       = GPIO_PIN_11 | GPIO_PIN_12;
    gpio_init.Mode      = GPIO_MODE_AF_PP;
    gpio_init.Pull      = GPIO_NOPULL;
    gpio_init.Speed     = GPIO_SPEED_FREQ_VERY_HIGH;
    gpio_init.Alternate = GPIO_AF10_OTG_FS;
    HAL_GPIO_Init(GPIOA, &gpio_init);

    return 0;
}

/* TinyUSB Descriptor Callbacks */

uint8_t const *tud_descriptor_device_cb(void) {
    return (const uint8_t *)&s_tud_desc_dev;
}

uint8_t const *tud_descriptor_configuration_cb(uint8_t index) {
    return s_tud_desc_configuration;
}

uint16_t const *tud_descriptor_string_cb(uint8_t index, uint16_t langid) {
    switch (index) {
        case APP_TUSB_STRING_ID_LANG:
            return s_tud_desc_string_lang;
            break;
        case APP_TUSB_STRING_ID_MFG:
            return s_tud_desc_string_mfg;
            break;
        case APP_TUSB_STRING_ID_PRODUCT:
            return s_tud_desc_string_product;
            break;
        case APP_TUSB_STRING_ID_SERIAL:
            return s_tud_desc_string_serial;
            break;
        default:
            return s_tud_desc_string_unknown;
            break;
    }
}

uint8_t const *tud_hid_descriptor_report_cb(uint8_t instance) {
    return s_tud_desc_hid_report;
}

uint16_t tud_hid_get_report_cb(uint8_t instance, uint8_t report_id, hid_report_type_t report_type, uint8_t *buffer,
                               uint16_t reqlen) {
    switch (report_id) {
        case REPORT_ID_LIGHTING_LAMP_ARRAY_ATTRIBUTES: {
            return app_lighting_get_array_attributes(buffer);
            break;
        }
        case REPORT_ID_LIGHTING_LAMP_ATTRIBUTES_RESPONSE: {
            return app_lighting_get_lamp_attributes(buffer);
            break;
        }
        default: {
            break;
        }
    }

    return 0;
}

void tud_hid_set_report_cb(uint8_t instance, uint8_t report_id, hid_report_type_t report_type, uint8_t const *buffer,
                           uint16_t bufsize) {
    switch (report_id) {
        case REPORT_ID_LIGHTING_LAMP_ATTRIBUTES_REQUEST: {
            app_lighting_set_lamp_attributes_req(buffer);
            break;
        }
        case REPORT_ID_LIGHTING_LAMP_MULTI_UPDATE: {
            app_lighting_set_lamp_multi_update(buffer);
            break;
        }
        case REPORT_ID_LIGHTING_LAMP_RANGE_UPDATE: {
            app_lighting_set_lamp_range_update(buffer);
        }
        case REPORT_ID_LIGHTING_LAMP_ARRAY_CONTROL: {
            app_lighting_set_lamp_array_control(buffer);
            break;
        }
        default: {
            break;
        }
    }
}